part of nueli.core.config;

InkConfig get $ink => InkConfig();

class InkConfig extends Config<InkConfig, Ink> {
  Color? _splash;
  Color? _hover;
  Color? _focus;
  Color? _overlay;
  Color? _highlight;
  double? _size;
  RadiiConfig? _radii;

  InkConfig() : super(Ink());

  InkConfig _config({
    Color? splash,
    Color? hover,
    Color? focus,
    Color? overlay,
    Color? highlight,
    double? radius,
    RadiiConfig? radii,
  }) {
    _splash = splash ?? _splash;
    _hover = hover ?? _hover;
    _focus = focus ?? _focus;
    _overlay = overlay ?? _overlay;
    _highlight = highlight ?? _highlight;
    _size = radius ?? _size;
    _radii = radii ?? _radii;
    return this;
  }

  InkConfig ripple(Color color, [int? alpha]) => splash(color, alpha).hover(color, alpha).overlay(color, alpha);
  InkConfig splash(Color color, [int? alpha]) => _config(splash: color.withAlpha(alpha ?? palette.tint.alpha));
  InkConfig hover(Color color, [int? alpha]) => _config(hover: color.withAlpha(alpha ?? palette.tint.alpha ~/ 2));
  InkConfig focus(Color color, [int? alpha]) => _config(focus: color.withAlpha(alpha ?? palette.tint.alpha ~/ 5));
  InkConfig overlay(Color color, [int? alpha]) => _config(overlay: color.withAlpha(alpha ?? palette.tint.alpha ~/ 2));
  InkConfig highlight(Color color, [int? alpha]) => _config(highlight: color.withAlpha(alpha ?? palette.tint.alpha ~/ 3));

  InkConfig radii(RadiiConfig radii) => _config(radii: radii);

  InkConfig withType(InkType? type) => this & type;
  InkConfig withTap(Function()? onTap) => this & onTap;
  InkConfig withLongPress(Function()? onLongPress) => this & Attr(#longpress)[onLongPress];

  @override
  InkConfig operator *(double factor) {
    final inkAp = _InkApprox(this);
    return _config(
      splash: inkAp.splash * factor,
      hover: inkAp.hover * factor,
      focus: inkAp.focus * factor,
      overlay: inkAp.overlay * factor,
      highlight: inkAp.highlight * factor,
      radius: inkAp.size * factor,
      radii: inkAp.radii * factor,
    );
  }

  @override
  InkConfig operator -(InkConfig operand) {
    final inkAp = _InkApprox(this);
    final inkAp2 = _InkApprox(operand);
    return _config(
      splash: inkAp.splash - inkAp2.splash,
      hover: inkAp.hover - inkAp2.hover,
      focus: inkAp.focus - inkAp2.focus,
      overlay: inkAp.overlay - inkAp2.overlay,
      highlight: inkAp.highlight - inkAp2.highlight,
      radius: inkAp.size - inkAp2.size,
      radii: inkAp.radii - inkAp2.radii,
    );
  }

  @override
  InkConfig operator +(InkConfig operand) {
    final inkAp = _InkApprox(this);
    final inkAp2 = _InkApprox(operand);
    return _config(
      splash: inkAp.splash + inkAp2.splash,
      hover: inkAp.hover + inkAp2.hover,
      focus: inkAp.focus + inkAp2.focus,
      overlay: inkAp.overlay + inkAp2.overlay,
      highlight: inkAp.highlight + inkAp2.highlight,
      radius: inkAp.size + inkAp2.size,
      radii: inkAp.radii + inkAp2.radii,
    );
  }

  @override
  InkConfig lerp(ink) {
    if (ink is InkConfig)
      _config(
        splash: ink._splash ?? _splash,
        hover: ink._hover ?? _hover,
        focus: ink._focus ?? _focus,
        overlay: ink._overlay ?? _overlay,
        highlight: ink._highlight ?? _highlight,
        radius: ink._size ?? _size,
        radii: ink._radii ?? _radii,
      );
    return super.lerp(ink);
  }

  Widget call() {
    final inkAp = _InkApprox(this);

    return data.type == InkType.response
        ? InkResponse(
            child: data.child,
            highlightColor: inkAp.overlay,
            overlayColor: MaterialStateProperty.all<Color>(inkAp.overlay),
            splashColor: inkAp.splash,
            hoverColor: inkAp.hover,
            focusColor: inkAp.focus,
            radius: max(Material.defaultSplashRadius, inkAp.size * 0.7),
            onTap: this.data.onTap,
            onLongPress: this.data.onLongPress,
          )
        : InkWell(
            child: data.child,
            borderRadius: inkAp.radii.toBorderRadius,
            splashColor: inkAp.splash,
            hoverColor: inkAp.hover,
            focusColor: inkAp.focus,
            overlayColor: MaterialStateProperty.all<Color>(inkAp.overlay),
            highlightColor: inkAp.highlight,
            radius: _size,
            onTap: this.data.onTap,
            onLongPress: this.data.onLongPress,
          );
  }
}

class Ink extends ConfigData<Ink> {
  InkType? type;
  Function()? onTap;
  Function()? onLongPress;
  Ink({Widget? child, this.type, this.onTap, this.onLongPress}) : super(child);

  Ink lerp(ink) {
    if (ink is InkType) type = ink;
    if (ink is Function()) onTap = ink;

    if (ink is Attr && ink is Function()) {
      if (ink.embedded == #tap) onTap = ink.value;
      if (ink.embedded == #longpress) onLongPress = ink.value;
    }

    if (ink is Ink) {
      type = ink.type;
      if (ink is Ink) onTap = ink.onTap;
      if (ink is Ink) onLongPress = ink.onLongPress;
    }
    return super.lerp(ink);
  }
}

class InkTween extends Tween<InkConfig> {
  InkTween({InkConfig? begin, InkConfig? end}) : super(begin: begin, end: end);

  @override
  InkConfig transform(double value) => (begin ?? $ink) + ((end ?? $ink) - (begin ?? $ink)) * value;
}

class InkTransition extends AnimatedWidget {
  final Widget? child;
  InkTransition({required Animation<InkConfig> ink, this.child}) : super(listenable: ink);
  Animation<InkConfig> get ink => listenable as Animation<InkConfig>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => ink.value.withChild(child)(),
    );
  }
}

enum InkType { response, well }

class _InkApprox {
  final InkConfig data;
  _InkApprox(this.data);

  Color get splash => data._splash ?? palette.accent.withAlpha(palette.tint.alpha);
  Color get hover => data._hover ?? palette.accent.withAlpha(palette.tint.alpha ~/ 2);
  Color get focus => data._focus ?? palette.accent.withAlpha(palette.tint.alpha ~/ 5);
  Color get overlay => data._overlay ?? palette.accent.withAlpha(palette.tint.alpha ~/ 2);
  Color get highlight => data._highlight ?? palette.accent.withAlpha(palette.tint.alpha ~/ 3);
  double get size => data._size ?? 200;
  RadiiConfig get radii => data._radii ?? $radii.zero;
}
