part of nueli.core.config;

TranxConfig get $tranx => TranxConfig();

class TranxConfig extends Config<TranxConfig, Tranx> {
  List<Tween> _tween = [];

  TranxConfig() : super(Tranx());

  TranxConfig _config({
    Tween<double>? scale,
    Tween<Offset>? translate,
    Tween<double>? height,
    Tween<double>? width,
    Tween<double>? rotation,
    Tween<double>? opacity,
    Tween<Config>? configTween,
  }) {
    if (scale != null) _tween.add(_Tween.tween(_TweenType.scale, scale));
    if (translate != null) _tween.add(_Tween.tween(_TweenType.translate, translate));
    if (height != null) _tween.add(_Tween.tween(_TweenType.height, height));
    if (width != null) _tween.add(_Tween.tween(_TweenType.width, width));
    if (rotation != null) _tween.add(_Tween.tween(_TweenType.rotation, rotation));
    if (opacity != null) _tween.add(_Tween.tween(_TweenType.opacity, opacity));
    if (configTween != null) _tween.add(configTween);
    return this;
  }

  TranxConfig translate(Offset begin, Offset end) => _config(translate: Tween(begin: begin, end: end));
  TranxConfig scale(double begin, double end, [Alignment alignment = Alignment.center]) => _config(scale: Tween(begin: begin, end: end));
  TranxConfig size(Size begin, Size end, [Alignment alignment = Alignment.center]) => size_vert(begin.width, end.width).size_horz(begin.height, end.height);
  TranxConfig size_vert(double begin, double end, [Alignment alignment = Alignment.center]) => _config(width: Tween(begin: begin, end: end)); // ignore: non_constant_identifier_names
  TranxConfig size_horz(double begin, double end, [Alignment alignment = Alignment.center]) => _config(height: Tween(begin: begin, end: end)); // ignore: non_constant_identifier_names

  TranxConfig rotate(double begin, double end) => _config(rotation: Tween(begin: begin, end: end));
  TranxConfig fade(double begin, double end) => _config(opacity: Tween(begin: begin, end: end));

  TranxConfig decor(DecorConfig begin, DecorConfig end) => _config(configTween: DecorTween(begin: begin, end: end));
  TranxConfig ink(InkConfig begin, InkConfig end) => _config(configTween: InkTween(begin: begin, end: end));
  TranxConfig media(PhotoConfig begin, PhotoConfig end) => _config(configTween: PhotoTween(begin: begin, end: end));
  TranxConfig radii(RadiiConfig begin, RadiiConfig end) => _config(configTween: RadiiTween(begin: begin, end: end));
  TranxConfig ace(AceConfig begin, AceConfig end) => _config(configTween: AceTween(begin: begin, end: end));
  TranxConfig txt(TxtConfig begin, TxtConfig end) => _config(configTween: TxtTween(begin: begin, end: end));
  TranxConfig x3D(X3DConfig begin, X3DConfig end) => _config(configTween: X3DTween(begin: begin, end: end));

  TranxConfig withAnimation(Animation<double>? animation) => this & animation;

  @override
  @deprecated
  TranxConfig operator *(double factor) => this;

  @override
  @deprecated
  TranxConfig operator +(TranxConfig operand) => this & operand;

  @override
  @deprecated
  TranxConfig operator -(TranxConfig operand) => this & operand;

  @override
  TranxConfig lerp(tranx) {
    if (tranx is TranxConfig) _tween.addAll(tranx._tween);
    return super.lerp(tranx);
  }

  Widget call() {
    Widget child = data.child ?? SizedBox();
    for (Tween tween in _tween) {
      if (data.animation == null) continue;
      if (tween is _Tween<double>) {
        if (tween.type == _TweenType.scale) child = ScaleTransition(scale: tween.animate(data.animation!), child: child);
        if (tween.type == _TweenType.height) child = SizeTransition(axis: Axis.vertical, sizeFactor: tween.animate(data.animation!), child: child);
        if (tween.type == _TweenType.width) child = SizeTransition(axis: Axis.horizontal, sizeFactor: tween.animate(data.animation!), child: child);
        if (tween.type == _TweenType.opacity) child = FadeTransition(opacity: tween.animate(data.animation!), child: child);
        if (tween.type == _TweenType.rotation) child = RotationTransition(turns: tween.animate(data.animation!), child: child);
      }
      if (tween is _Tween<Offset>) {
        if (tween.type == _TweenType.translate) child = SlideTransition(position: tween.animate(data.animation!), child: child);
      }
      if (tween is Tween<Config>) {
        if (tween is DecorTween) child = DecorTransition(decor: tween.animate(data.animation!), child: child);
        if (tween is InkTween) child = InkTransition(ink: tween.animate(data.animation!), child: child);
        if (tween is PhotoTween) child = PhotoTransition(photo: tween.animate(data.animation!), child: child);
        if (tween is RadiiTween) child = RadiiTransition(radii: tween.animate(data.animation!), child: child);
        if (tween is AceTween) child = AceTransition(ace: tween.animate(data.animation!), child: child);
        if (tween is TxtTween) child = TxtTransition(txt: tween.animate(data.animation!), child: child);
        if (tween is X3DTween) child = X3DTransition(x3D: tween.animate(data.animation!), child: child);
      }
      return child;
    }
    return child;
  }
}

class Tranx extends ConfigData<Tranx> {
  Animation<double>? animation;
  Tranx({Widget? child, this.animation}) : super(child);

  Tranx lerp(data) {
    if (data is Animation<double>) animation = data;
    if (data is Tranx) animation = data.animation;
    return super.lerp(data);
  }
}

enum _TweenType { scale, translate, height, width, rotation, opacity }

class _Tween<T> extends Tween<T> {
  final _TweenType type;
  _Tween(this.type, {T? begin, T? end}) : super(begin: begin, end: end);
  _Tween.tween(this.type, Tween<T> tween) : super(begin: tween.begin, end: tween.end);
}
