part of nueli.core.config;

TxtConfig get $txt => TxtConfig();

/// Text config
class TxtConfig extends Config<TxtConfig, Txt> {
  List<TxtConfig> _children = [];
  Color? _color, _backgroundColor;
  double? _size;
  AceConfig? _shadow;
  InkConfig? _ink;

  TxtConfig() : super(Txt());

  TxtConfig _config({
    Color? color,
    Color? background,
    double? size,
    AceConfig? shadow,
    InkConfig? ink,
  }) {
    _color = color ?? _color;
    _backgroundColor = background ?? _backgroundColor;
    _size = size ?? _size;
    _shadow = shadow ?? _shadow;
    _ink = ink ?? _ink;
    return this;
  }

  TxtConfig color(Color color) => _config(color: color);
  TxtConfig get color_bg => _config(color: palette.background); // ignore: non_constant_identifier_names
  TxtConfig get color_text => _config(color: palette.text); // ignore: non_constant_identifier_names
  TxtConfig get color_hint => _config(color: palette.text.withAlpha(150)); // ignore: non_constant_identifier_names

  TxtConfig size(double size) => _config(size: size);
  TxtConfig get size_10 => size(10);
  TxtConfig get size_12 => size(12);
  TxtConfig get size_14 => size(14);
  TxtConfig get size_16 => size(16);
  TxtConfig get size_18 => size(18);
  TxtConfig get size_24 => size(24);

  TxtConfig shadow(AceConfig shadow) => _config(shadow: shadow);

  TxtConfig ink(InkConfig react) => _config(ink: react);

  TxtConfig withText(String? text) => this & text;
  TxtConfig withWeight(FontWeight? weight) => this & weight;
  TxtConfig withDecoration(TextDecoration? decoration) => this & decoration;
  TxtConfig withBaseline(TextBaseline? baseline) => this & baseline;
  TxtConfig withOverflow(TextOverflow? overflow) => this & overflow;
  TxtConfig withStyle(FontStyle? style) => this & style;
  TxtConfig withAlign(TextAlign? align) => this & align;

  @override
  TxtConfig operator *(double factor) {
    final txtAp = _TxtApprox(this);
    return _config(
      color: txtAp.color * factor,
      background: txtAp.backgroundColor * factor,
      size: txtAp.size * factor,
      shadow: txtAp.shadow * factor,
      ink: txtAp.ink * factor,
    );
  }

  @override
  TxtConfig operator -(TxtConfig operand) {
    final txtAp = _TxtApprox(this);
    final txtAp2 = _TxtApprox(operand);
    return _config(
      color: txtAp.color - txtAp2.color,
      background: txtAp.backgroundColor - txtAp2.backgroundColor,
      size: txtAp.size - txtAp2.size,
      shadow: txtAp.shadow - txtAp2.shadow,
      ink: txtAp.ink - txtAp2.ink,
    );
  }

  @override
  TxtConfig operator +(TxtConfig operand) {
    final txtAp = _TxtApprox(this);
    final txtAp2 = _TxtApprox(operand);
    return _config(
      color: txtAp.color + txtAp2.color,
      background: txtAp.backgroundColor + txtAp2.backgroundColor,
      size: txtAp.size + txtAp2.size,
      shadow: txtAp.shadow + txtAp2.shadow,
      ink: txtAp.ink + txtAp2.ink,
    );
  }

  TxtConfig lerp(txt) {
    if (txt is TxtConfig)
      _config(
        color: txt._color ?? _color,
        background: txt._backgroundColor ?? _backgroundColor,
        size: txt._size ?? _size,
        shadow: txt._shadow ?? _shadow,
        ink: txt._ink ?? _ink,
      );
    return super.lerp(txt);
  }

  TxtConfig save() => TxtConfig().._children.addAll([...this._children, this.._children = []]);

  TxtConfig inline(List<TxtConfig> children) => this.._children = children;

  TxtConfig highlight(String charseq, TxtConfig config) {
    List<String> children = data.text.split(RegExp(charseq));
    List<Match> matches = RegExp(charseq).allMatches(data.text).toList();
    List<TxtConfig> configList = [];
    for (int i = 1; i < children.length; i++) {
      String match = matches[i - 1].group(0) ?? '';
      configList.add(config & match);
      configList.add(this & children[i]);
    }
    return (this & children.first).inline(configList);
  }

  String toString() => data.text;

  TextStyle get toTextStyle {
    final txtAp = _TxtApprox(this);
    return TextStyle(
      color: txtAp.color,
      decoration: TextDecoration.combine(data.decorations),
      fontSize: txtAp.size,
      shadows: txtAp.shadow.toBoxShadowDeep,
      fontStyle: data.style,
      fontWeight: data.weight,
      backgroundColor: txtAp.backgroundColor,
      textBaseline: TextBaseline.alphabetic,
    );
  }

  GestureRecognizer? get toRecognizer {
    final txtAp = _TxtApprox(this);
    if (txtAp.ink.data.onTap != null) return TapGestureRecognizer()..onTap = () => (txtAp.ink.data.onTap ?? () {})();
    if (txtAp.ink.data.onLongPress != null) return LongPressGestureRecognizer()..onLongPress = () => (txtAp.ink.data.onLongPress ?? () {})();
    return null;
  }

  Widget call() {
    final txtAp = _TxtApprox(this);

    if (_children.isNotEmpty) {
      List<TextSpan> spans = [];
      for (TxtConfig config in _children)
        spans.add(TextSpan(
          text: (this | config).data.text,
          style: (this | config).toTextStyle,
          recognizer: (this | config).toRecognizer,
        ));
      return Container(
        width: data.fill ? double.infinity : null,
        child: Text.rich(
            TextSpan(
              text: data.text,
              style: toTextStyle,
              children: spans,
              recognizer: toRecognizer,
            ), // break
            textAlign: data.align,
            maxLines: data.maxLines,
            overflow: data.overflow,
            style: toTextStyle),
      );
    }

    final text = Container(
      width: data.fill ? double.infinity : null,
      child: Text(data.text, // break
          textAlign: data.align,
          maxLines: data.maxLines,
          overflow: data.overflow,
          style: toTextStyle),
    );
    if (_ink == null) return text;
    return (txtAp.ink & text)();
  }
}

class Txt extends ConfigData<Txt> {
  bool fill;
  String text;
  int? maxLines;
  FontWeight weight;
  List<TextDecoration> decorations;
  TextBaseline? baseline;
  TextOverflow? overflow;
  FontStyle? style;
  TextAlign align;

  Txt({
    this.fill = false,
    this.text = '',
    this.maxLines,
    this.weight = FontWeight.normal,
    this.decorations = const [],
    this.baseline,
    this.overflow,
    this.style,
    this.align = TextAlign.start,
  }) : super(SizedBox());

  Txt lerp(data) {
    if (data is String) text = data;
    if (data is FontWeight) weight = data;
    if (data is List<TextDecoration>) decorations = data;
    if (data is TextDecoration) decorations.add(data);
    if (data is TextBaseline) baseline = data;
    if (data is TextOverflow) overflow = data;
    if (data is FontStyle) style = data;
    if (data is TextAlign) align = data;

    if (data is Txt) text = data.text;
    if (data is Txt) maxLines = data.maxLines;
    if (data is Txt) weight = data.weight;
    if (data is Txt) decorations = data.decorations;
    if (data is Txt) baseline = data.baseline;
    if (data is Txt) overflow = data.overflow;
    if (data is Txt) style = data.style;
    if (data is Txt) align = data.align;
    return super.lerp(data);
  }
}

class TxtInk extends Ink {
  final Function(TxtConfig text)? onTextTap;
  final Function(TxtConfig text)? onTextLongPress;
  TxtInk({
    Function()? onTap,
    Function()? onLongPress,
    this.onTextTap,
    this.onTextLongPress,
  }) : super(type: InkType.well, onTap: onTap, onLongPress: onLongPress);
}

class TxtTween extends Tween<TxtConfig> {
  TxtTween({TxtConfig? begin, TxtConfig? end}) : super(begin: begin, end: end);

  @override
  TxtConfig transform(double value) => (begin ?? $txt) + ((end ?? $txt) - (begin ?? $txt)) * value;
}

class TxtTransition extends AnimatedWidget {
  final Widget? child;
  TxtTransition({required Animation<TxtConfig> txt, this.child}) : super(listenable: txt);
  Animation<TxtConfig> get txt => listenable as Animation<TxtConfig>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => txt.value.withChild(child)(),
    );
  }
}

class _TxtApprox {
  final TxtConfig txt;
  _TxtApprox(this.txt);

  Color get color => txt._color ?? palette.text;
  Color get backgroundColor => txt._backgroundColor ?? palette.none;
  double get size => txt._size ?? 12;
  AceConfig get shadow => txt._shadow ?? $ace;
  InkConfig get ink => txt._ink ?? $ink;
}
