part of nueli.core.config;

X3DConfig get $x3D => X3DConfig();

class X3DConfig extends Config<X3DConfig, X3D> {
  Alignment? _alignment;
  List<_TransformData> _transforms = [];

  X3DConfig() : super(X3D());

  X3DConfig _config({
    Alignment? alignment,
    List<_TransformData> transforms = const [],
  }) {
    _alignment = alignment ?? _alignment;
    _transforms = _transforms..addAll(transforms);
    return this;
  }

  X3DConfig align(Alignment alignment) => _config(alignment: alignment);

  X3DConfig skew({double x = 0, double y = 0, double z = 0}) => (skew_x(x)..skew_y(y))..rotate(z);
  X3DConfig skew_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..rotateX(d))]); // ignore: non_constant_identifier_names
  X3DConfig skew_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..rotateY(d))]); // ignore: non_constant_identifier_names
  X3DConfig rotate(double z) => _config(transforms: [_TransformData(z, (t, d) => t..rotateZ(d))]); // ignore: non_constant_identifier_names

  X3DConfig zoom(double zoom) => _config(transforms: [_TransformData(zoom, (t, d) => t..rotateZ(d))]);

  X3DConfig pan({double x = 0, double y = 0, double z = 0}) => (pan_x(x)..pan_y(y))..pan_z(z);
  X3DConfig pan_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..translate(d))]); // ignore: non_constant_identifier_names
  X3DConfig pan_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..translate(0, d))]); // ignore: non_constant_identifier_names
  X3DConfig pan_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..translate(0, 0, d))]); // ignore: non_constant_identifier_names

  X3DConfig shift({double x = 0, double y = 0, double z = 0}) => (shift_x(x)..shift_y(y))..shift_z(z);
  X3DConfig shift_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..leftTranslate(d))]); // ignore: non_constant_identifier_names
  X3DConfig shift_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..leftTranslate(0, d))]); // ignore: non_constant_identifier_names
  X3DConfig shift_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..leftTranslate(0, 0, d))]); // ignore: non_constant_identifier_names

  X3DConfig scale({double x = 0, double y = 0, double z = 0}) => (scale_x(x)..scale_y(y))..scale_z(z);
  X3DConfig scale_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..scale(d))]); // ignore: non_constant_identifier_names
  X3DConfig scale_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..scale(0, d))]); // ignore: non_constant_identifier_names
  X3DConfig scale_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..scale(0, 0, d))]); // ignore: non_constant_identifier_names

  X3DConfig withBase(Matrix4? base) => this & base;

  @override
  X3DConfig operator *(double factor) {
    final x3DAp = _X3DApprox(this);
    return _config(
      alignment: x3DAp.alignment * factor,
      transforms: _transforms.map((e) => e * factor).toList(),
    );
  }

  @override
  X3DConfig operator -(X3DConfig operand) {
    final x3DAp = _X3DApprox(this);
    final x3DAp2 = _X3DApprox(operand);
    List<_TransformData> transforms = _transforms.minus(operand._transforms);
    return X3DConfig() & data
      .._config(alignment: x3DAp.alignment - x3DAp2.alignment, transforms: transforms);
  }

  @override
  X3DConfig operator +(X3DConfig operand) {
    final x3DAp = _X3DApprox(this);
    final x3DAp2 = _X3DApprox(operand);
    List<_TransformData> transforms = _transforms.plus(operand._transforms);
    return _config(alignment: x3DAp.alignment + x3DAp2.alignment, transforms: transforms);
  }

  @override
  X3DConfig lerp(matrix) {
    if (matrix is X3DConfig)
      _config(
        alignment: matrix._alignment ?? _alignment,
        transforms: matrix._transforms + _transforms,
      );
    return super.lerp(matrix);
  }

  Matrix4 get toMatrix4 {
    Matrix4 mat = data.base;
    for (_TransformData transform in _transforms) mat = transform(mat);
    return mat;
  }

  Widget call() {
    _alignment = _alignment ?? Alignment.center;
    return Transform(child: data.child, alignment: _alignment, transform: toMatrix4);
  }
}

class X3D extends ConfigData<X3D> {
  late Matrix4 base;

  X3D({Widget? child, Matrix4? base}) : super(child) {
    base = Matrix4.identity()..setEntry(3, 2, 0.001);
  }

  X3D lerp(data) {
    if (data is Matrix4) base = data;

    if (data is X3D) base = data.base;
    return super.lerp(data);
  }
}

class X3DTween extends Tween<X3DConfig> {
  X3DTween({X3DConfig? begin, X3DConfig? end}) : super(begin: begin, end: end);

  @override
  X3DConfig transform(double value) => (begin ?? $x3D) + ((end ?? $x3D) - (begin ?? $x3D)) * value;
}

class X3DTransition extends AnimatedWidget {
  final Widget? child;
  X3DTransition({required Animation<X3DConfig> x3D, this.child}) : super(listenable: x3D);
  Animation<X3DConfig> get x3D => listenable as Animation<X3DConfig>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => x3D.value.withChild(child)(),
    );
  }
}

class _X3DApprox {
  final X3DConfig x3D;
  _X3DApprox(this.x3D);

  Alignment get alignment => x3D._alignment ?? Alignment.center;
  List<_TransformData> get transforms => x3D._transforms;
}

class _TransformData with Subtract<_TransformData>, Add<_TransformData> {
  double value;
  Matrix4 Function(Matrix4, double) transform;
  _TransformData(this.value, this.transform);
  Matrix4 call(Matrix4 matrix) => transform(matrix, value);

  _TransformData operator -(_TransformData element) {
    return _TransformData(value - element.value, (mat, d) => transform(element.transform(mat, d), d));
  }

  @override
  _TransformData operator +(_TransformData element) {
    return _TransformData(value + element.value, (mat, d) => transform(element.transform(mat, d), d));
  }

  _TransformData operator *(double factor) {
    return _TransformData(value + factor, (mat, d) => transform(mat, d));
  }
}
