library nueli.core.event;

import 'dart:async';

import 'data.dart' show $, Ref, Add;

// Events
typedef Construct<T> = T Function(); // Interlaced with setters
typedef Init<T> = dynamic Function(T data);
typedef Modify<T> = T Function(T data);
typedef Use<T> = void Function(T data); // Interlaced with getters
typedef Assign<K, T> = void Function(K key, T data);
typedef Query<K, T> = T Function(K key, [T fallback]);
typedef Convert<D, T> = D Function(T data);
typedef Test<T> = bool Function(T data);

// Async Events
typedef AsyncConstruct<T> = Future<T> Function();
typedef AsyncInit<T> = Future<dynamic> Function(T data);
typedef AsyncModify<T> = Future<T> Function(T data);
typedef AsyncUse<T> = Future<void> Function(T data);
typedef AsyncAssign<K, T> = Future<void> Function(K key, T data);
typedef AsyncQuery<K, T> = Future<T> Function(K key, [T fallback]);
typedef AsyncConvert<D, T> = Future<D> Function(T data);
typedef AsyncTest<T> = Future<bool> Function(T data);

/// Event exception
extension EventException on Exception {
  get msg => this.toString().replaceFirst('Exception: ', '');
}

/// Event Exception Handler
T? handle<T>(
  T? Function() event, {
  T? Function(String msg)? alt,
  T? Function(T? condition, bool success)? ult,
}) {
  T? _alt(String e) => (alt ?? (e) => null)(e);
  T? _ult(T? c, bool s) => (ult ?? (c, s) => c)(c, s);
  try {
    return _ult(event(), true);
  } catch (e) {
    try {
      return _ult(_alt((e as Exception).msg), false);
    } catch (e) {
      return _ult(_alt(''), false);
    }
  }
}

/// Async Event Exception Handler
Future<T?> asyncHandle<T>(
  FutureOr<T?> Function() task, {
  T? Function(String msg)? alt,
  FutureOr<T?> Function(FutureOr<T?> condition, bool success)? ult,
}) async {
  T? _alt(String e) => (alt ?? (e) => null)(e);
  FutureOr<T?> _ult(FutureOr<T?> c, bool s) => (ult ?? (c, s) async => await c)(c, s);
  try {
    return await _ult(await task(), true);
  } catch (e) {
    try {
      return _ult(_alt((e as Exception).msg), false);
    } catch (e) {
      return _ult(_alt(''), false);
    }
  }
}

/// Shortcut to begin an event Algorithm
Algorithm get algog => Algorithm();

/// Event Algorithm <br>
/// <span color = '#99aaff'> Begin an algorithm</span> | [ag & async {task}] <br>
/// <span color = '#99aaff'> Create an empty algorithm<span> | [ag & null ] <br>
class Algorithm<T> with Add<Algorithm<T>> {
  final Ref<List<Future<T> Function()>> _flow = $([]);

  Algorithm([List<Future<T> Function()>? flow]) {
    _flow.data = (flow ?? []).toList();
  }

  Algorithm<T> operator &(Future<T> Function()? task) {
    if (task == null) {
      return Algorithm(_flow.data);
    } else
      return Algorithm<T>([...(_flow.data ?? []), task]);
  }

  @override
  Algorithm<T> operator +(Algorithm<T>? algorithm) {
    if (algorithm == null) {
      return Algorithm(_flow.data);
    } else {
      return Algorithm([...(_flow.data ?? []), ...(algorithm._flow.data ?? [])]);
    }
  }

  Future<T?> call({bool reversed = false}) async {
    T? result;
    for (Future<T> Function() task in (_flow.data ?? [])) {
      await task().then((value) => result = value);
    }
    return result;
  }
}
